from dunia.playwright.browser import (
    PlaywrightBrowser,
    PlaywrightBrowserWithLogin,
    PlaywrightBrowserWithoutLogin,
    PlaywrightPersistentBrowser,
)
from dunia.playwright.page import PlaywrightElementHandle, PlaywrightPage


__all__ = [
    "PlaywrightBrowser",
    "PlaywrightPersistentBrowser",
    "PlaywrightBrowserWithLogin",
    "PlaywrightBrowserWithoutLogin",
    "PlaywrightPage",
    "PlaywrightElementHandle",
]



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Protocol, cast

import playwright.async_api as playwright

from dunia.browser import Browser
from dunia.error import BrowserNotInitialized
from dunia.log import info
from dunia.login import Login
from dunia.playwright.page import PlaywrightPage
from dunia.request import Request

if TYPE_CHECKING:
    from pathlib import Path

    from dunia.browser import BrowserConfig
    from dunia.login import LoginInfo


class PlaywrightBrowser(Browser, Protocol):
    async def new_page(self) -> PlaywrightPage:
        ...

    @property
    def context(self) -> playwright.BrowserContext:
        ...


@dataclass(slots=True, kw_only=True)
class PlaywrightPersistentBrowser:
    """
    This class contains all the possible data/state required for initialization of Playwright browser (Chrome)
    """

    browser_config: BrowserConfig
    playwright: playwright.Playwright


@dataclass(slots=True, kw_only=True)
class PlaywrightBrowserWithLogin(PlaywrightPersistentBrowser):
    """
    Use this class when the market requires login for crawling the information
    """

    login_info: LoginInfo
    __browser_context: playwright.BrowserContext | None = field(
        default=None,
        init=False,
        repr=False,
    )

    async def login(self) -> PlaywrightBrowser:
        login = Login(self.login_info)
        self.__browser_context = await create_playwright_persistent_browser(self)
        await login(self)
        return self

    # ? We only need new_page() from playwright's BrowserContext to make this class API compatible with all the existing markets because we can then simply use PlaywrightBrowser instead of playwright's BrowserContext
    async def new_page(self) -> PlaywrightPage:
        if not self.__browser_context:
            raise BrowserNotInitialized("Please call login() first")

        return PlaywrightPage(await self.context.new_page())

    @property
    def context(self) -> playwright.BrowserContext:
        if not self.__browser_context:
            raise BrowserNotInitialized("Please call login() first")

        return self.__browser_context

    @property
    def request(self) -> Request:
        return cast(Request, self.context.request)


@dataclass(slots=True, kw_only=True)
class PlaywrightBrowserWithoutLogin(PlaywrightPersistentBrowser):
    """
    Use this class when the market doesn't require a login for crawling the information
    """

    __browser_context: playwright.BrowserContext | None = field(
        default=None,
        init=False,
        repr=False,
    )

    async def create(self) -> PlaywrightBrowser:
        self.__browser_context = await create_playwright_persistent_browser(self)
        return self

    # ? We only need new_page() from playwright's BrowserContext to make this class API compatible with all the existing markets because we can then simply use PlaywrightBrowser instead of playwright's BrowserContext
    async def new_page(self) -> PlaywrightPage:
        if not self.__browser_context:
            raise BrowserNotInitialized("Please call create() first")

        return PlaywrightPage(await self.__browser_context.new_page())

    @property
    def context(self) -> playwright.BrowserContext:
        if not self.__browser_context:
            raise BrowserNotInitialized("Please call create() first")

        return self.__browser_context

    @property
    def request(self) -> Request:
        return cast(Request, self.context.request)


async def create_playwright_persistent_browser(
    persistent_browser: PlaywrightPersistentBrowser,
) -> playwright.BrowserContext:
    """
    This creates the browser that saves the cache of visited pages in cache directory configured in BrowserConfig
    """
    if persistent_browser.browser_config.user_data_dir:
        info(
            f"Browser cache directory: <blue>{persistent_browser.browser_config.user_data_dir}</>"
        )
    browser_args: dict[
        str,
        Path
        | str
        | bool
        | list[str]
        | int
        | playwright.ViewportSize
        | playwright.ProxySettings,
    ] = dict(
        user_data_dir=persistent_browser.browser_config.user_data_dir,
        headless=persistent_browser.browser_config.headless,
        channel=persistent_browser.browser_config.channel,
        locale=persistent_browser.browser_config.locale,
        accept_downloads=persistent_browser.browser_config.accept_downloads,
        devtools=persistent_browser.browser_config.devtools,
    )

    if persistent_browser.browser_config.slow_mo:
        browser_args["slow_mo"] = persistent_browser.browser_config.slow_mo

    if persistent_browser.browser_config.viewport:
        browser_args["viewport"] = persistent_browser.browser_config.viewport
    else:
        browser_args["no_viewport"] = True

    if persistent_browser.browser_config.proxy:
        browser_args["proxy"] = persistent_browser.browser_config.proxy

    if persistent_browser.browser_config.browser == "chromium":
        browser = await persistent_browser.playwright.chromium.launch_persistent_context(**browser_args)  # type: ignore
    else:
        browser = await persistent_browser.playwright.firefox.launch_persistent_context(**browser_args)  # type: ignore

    browser.set_default_navigation_timeout(
        persistent_browser.browser_config.default_navigation_timeout
    )
    browser.set_default_timeout(persistent_browser.browser_config.default_timeout)

    return browser


async def close_first_blank_page(browser: PlaywrightBrowser):
    if len(browser.context.pages) > 1 and "about:blank" in browser.context.pages[0].url:
        await browser.context.pages[0].close()


async def main_page_only(browser: PlaywrightBrowser) -> PlaywrightPage:
    new_page = await browser.new_page()
    all_pages = browser.context.pages
    if len(all_pages) > 1:
        for page in all_pages[:-1]:
            await page.close()
    return new_page



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing_extensions import Self


class Element(Protocol):
    async def query_selector(self, selector: str) -> Self | None:
        ...

    async def query_selector_all(self, selector: str) -> list[Self]:
        ...

    async def text_content(self) -> str | None:
        ...

    async def get_attribute(self, name: str) -> str | None:
        ...


class Node(Element, Protocol):
    async def click(self, *, timeout: int | None = None) -> None:
        ...

    async def select_option(
        self,
        value: str | list[str] | None = None,
        *,
        label: str | list[str] | None = None,
        timeout: int | None = None,
    ) -> list[str]:
        ...

    async def scroll_into_view_if_needed(self, *, timeout: int | None = None) -> None:
        ...

    async def focus(self) -> None:
        ...

    async def is_visible(self) -> bool:
        ...

    async def is_enabled(self) -> bool:
        ...

    async def fill(self, value: str, *, timeout: int | None = None) -> None:
        ...

    async def type(self, text: str, *, timeout: int | None = None) -> None:
        ...

    async def press(self, key: str, *, timeout: int | None = None) -> None:
        ...

    async def inner_html(self) -> str:
        ...


# ? Aliases
Fragment = Element
ElementHandle = Node



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING

from dunia.modest._core import css, css_first

if TYPE_CHECKING:
    from selectolax.parser import HTMLParser, Node
    from typing_extensions import Self

    from dunia.element import Element


@dataclass(slots=True, frozen=True)
class ModestDocument:
    html_element: HTMLParser

    async def query_selector(self, selector: str) -> Element | None:
        if html_element := await css_first(self.html_element, selector):
            return ModestElement(html_element)  # type: ignore

        return None

    async def query_selector_all(self, selector: str) -> list[Element]:
        return [
            ModestElement(html_element)
            for html_element in await css(self.html_element, selector)
        ]

    async def text_content(
        self, selector: str, *, timeout: int | None = None
    ) -> str | None:
        if html_element := await css_first(self.html_element, selector):
            return html_element.text(deep=True)  # type: ignore

        return None

    async def inner_text(
        self, selector: str, *, timeout: int | None = None
    ) -> str | None:
        if html_element := await css_first(self.html_element, selector):
            return html_element.text(deep=False)  # type: ignore

        return None

    async def get_attribute(
        self, selector: str, name: str, *, timeout: int | None = None
    ) -> str | None:
        if html_element := await css_first(self.html_element, selector):
            return html_element.attrs.sget(name, None)  # type: ignore

        return None


@dataclass(slots=True, frozen=True)
class ModestElement:
    html_element: Node

    async def query_selector(self, selector: str) -> Self | None:
        if html_element := await css_first(self.html_element, selector):
            return ModestElement(html_element)  # type: ignore

        return None

    async def query_selector_all(self, selector: str) -> list[Self]:
        return [
            ModestElement(html_element)
            for html_element in await css(self.html_element, selector)
        ]

    async def text_content(self) -> str | None:
        if text := await asyncio.to_thread(self.html_element.text):
            return text

        return None

    async def get_attribute(self, name: str) -> str | None:
        return self.html_element.attrs.sget(name, default=None)



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Protocol, TypeVar, runtime_checkable

from dunia.element import Element, Node

if TYPE_CHECKING:
    import asyncio
    from pathlib import Path
    from types import NoneType
    from typing import TYPE_CHECKING, Any, Literal

ReturnSelectorType = TypeVar("ReturnSelectorType")
T = TypeVar("T")


class DocumentTextSelector(Protocol):
    async def text_content(
        self, selector: str, *, timeout: int | None = None
    ) -> str | None:
        ...

    async def inner_text(
        self, selector: str, *, timeout: int | None = None
    ) -> str | None:
        ...


class DocumentAttributeSelector(Protocol):
    async def get_attribute(
        self, selector: str, name: str, *, timeout: int | None = None
    ) -> str | None:
        ...


class QuerySelector(
    Protocol[ReturnSelectorType],
):
    async def query_selector(self, selector: str) -> ReturnSelectorType | None:
        ...

    async def query_selector_all(self, selector: str) -> list[ReturnSelectorType]:
        ...


@runtime_checkable
class Document(
    QuerySelector[Element], DocumentTextSelector, DocumentAttributeSelector, Protocol
):
    pass


@runtime_checkable
class Page(
    QuerySelector[Node], DocumentTextSelector, DocumentAttributeSelector, Protocol
):
    async def click(self, selector: str, *, timeout: int | None = None) -> None:
        ...

    async def press(
        self, selector: str, key: str, *, timeout: int | None = None
    ) -> None:
        ...

    async def fill(
        self, selector: str, value: str, *, timeout: int | None = None
    ) -> None:
        ...

    async def evaluate(self, expression: str) -> Any:
        ...

    async def type(
        self, selector: str, text: str, *, timeout: int | None = None
    ) -> None:
        ...

    async def inner_html(
        self, selector: str, *, timeout: int | None = None
    ) -> str | None:
        ...

    @property
    def mouse(self) -> Any:
        ...

    async def content(self) -> str:
        ...

    async def set_content(
        self,
        html: str,
        *,
        timeout: int | None = None,
        wait_until: Literal[
            "commit", "domcontentloaded", "load", "networkidle"
        ] = "load",
    ) -> None:
        ...

    async def goto(
        self,
        url: str,
        *,
        timeout: int | None = None,
        wait_until: Literal[
            "commit", "domcontentloaded", "load", "networkidle"
        ] = "load",
    ) -> None:
        ...

    async def close(self) -> None:
        ...

    async def pause(self) -> None:
        ...

    @property
    def url(self) -> str:
        ...

    async def screenshot(
        self, path: str, *, full_page: bool = False, timeout: int | None = None
    ) -> bytes | None:
        ...

    async def wait_for_load_state(
        self,
        state: Literal["domcontentloaded", "load", "networkidle"],
        *,
        timeout: int | None = None,
    ) -> None:
        ...

    async def wait_for_selector(
        self,
        selector: str,
        *,
        timeout: int | None = None,
        state: Literal["attached", "detached", "hidden", "visible"] | None = None,
    ) -> Node | None:
        ...

    async def wait_for_timeout(self, timeout: int) -> None:
        ...

    async def select_option(
        self,
        selector: str,
        value: str | list[str] | None = None,
        *,
        label: str | list[str] | None = None,
        timeout: int | None = None,
    ) -> list[str]:
        ...

    def frame(self, name: str) -> Frame | None:
        ...

    def expect_navigation(
        self, *, timeout: int | None = None
    ) -> AsyncEventContextManager[None]:
        ...

    def expect_download(
        self, *, timeout: int | None = None
    ) -> AsyncEventContextManager[Download]:
        ...

    async def check(self, selector: str, *, timeout: int | None = None) -> None:
        ...

    async def reload(self, *, timeout: int | None = None) -> None:
        ...


# ? Interfaces taken from Playwright's code
class AsyncEventInfo(Generic[T]):
    def __init__(self, future: asyncio.Future[T]) -> None:
        ...

    @property
    async def value(self) -> T:
        ...

    def is_done(self) -> bool:
        ...


class AsyncEventContextManager(Generic[T]):
    def __init__(self, future: asyncio.Future[T]) -> None:
        ...

    async def __aenter__(self) -> AsyncEventInfo[T]:
        ...

    async def __aexit__(self, *args: Any) -> None:
        ...


# ? We only need these methods from Playwright's Download
class Download(Protocol):
    @property
    def suggested_filename(self) -> str:
        ...

    async def save_as(self, path: str | Path) -> NoneType:
        ...


class FileChooser(Protocol):
    @property
    def page(self) -> Page:
        ...

    @property
    def element(self) -> Node:
        ...

    def is_multiple(self) -> bool:
        ...

    async def set_files(
        self,
        files: str | Path | list[str | Path],
        *,
        timeout: float | None = None,
        no_wait_after: bool | None = None,
    ) -> NoneType:
        ...


# ? We only need these methods from Playwright's Frame
class Frame(Protocol):
    async def query_selector(self, selector: str) -> Node | None:
        ...

    async def query_selector_all(self, selector: str) -> list[Node]:
        ...
